"""File with weigthed regression model algorithm."""
# Third Party Library
from pamm.algorithms.regression.regression_model import RegressionAlgorithm

import numpy as np

class ModifyRegressionAlgorithm(RegressionAlgorithm):
    
    def _add_bias(self, x: np.ndarray, q: np.ndarray, b_apr: np.ndarray) -> tuple[np.ndarray]:
        if self._fit_bias:
            return (np.hstack([np.ones((x.shape[0], 1)),x]),
            np.diag(np.hstack([0,np.diag(q)])),
            np.vstack([0,b_apr]))
        return x , q, b_apr
    
    def fit(self,
            x: np.ndarray,
            y: np.ndarray,
            w: np.ndarray,
            q: np.ndarray,
            b_apr: np.ndarray
            ) -> None:
        """Fit model.

        Args:
            x (np.ndarray): matrix of regressors
            y (np.ndarray): matrix a teacher
            w (np.ndarray): matrix of weights for data
            q (np.ndarray): matrix of weights for beta
            b_apr (np.ndarray): knowns beta coefficients
        """
                   
        x, q, b_apr = self._add_bias(x, q, b_apr)
        tmp =  b_apr + (np.linalg.inv(q + x.T @ w @ x) @ x.T @ w) @ (y - x @ b_apr)
        self._beta = tmp[int(self._fit_bias):]
        self._bias  = tmp[0][0] * int(self._fit_bias)


"""Module with statistick methods between two time-series."""

# Standart Library
import numpy as np

def mse(major_ts: np.array, minor_ts: np.array) -> float:
    """Calculate mean squared error between 2 metricks.

    Args:
        major_ts (np.array): original data vector.
        minor_ts (np.array): predicted data vector.

    Returns:
        float: mean squared  error.
    """
    return ((major_ts-minor_ts)**2).mean()


def rmse(major_ts: np.array, minor_ts: np.array) -> float:
    """Calculate root mean squared error between 2 metricks.

    Args:
        major_ts (np.array): original data vector.
        minor_ts (np.array): predicted data vector.

    Returns:
        float: root mean squared error.
    """
    return np.sqrt(mse(major_ts, minor_ts))


def mae(major_ts: np.array, minor_ts: np.array) -> float:
    """Calculate mean absolute error between 2 metricks.

    Args:
        major_ts (np.array): original data vector.
        minor_ts (np.array): predicted data vector.

    Returns:
        float: absolute squared error.
    """
    return np.mean(np.abs(major_ts - minor_ts))


def mspe(major_ts: np.array, minor_ts: np.array) -> float:
    """Calculate mean squared percentable error between 2 metricks.

    Args:
        major_ts (np.array): original data vector.
        minor_ts (np.array): predicted data vector.

    Returns:
        float: mean squared percentable error.
    """
    
    return 100 * (((major_ts-minor_ts)/major_ts)**2).mean()


def mape(major_ts: np.array, minor_ts: np.array) -> float:
    """Calculate mean absolute percecentable error between 2 metricks.

    Args:
        major_ts (np.array): original data vector.
        minor_ts (np.array): predicted data vector.

    Returns:
        float: mean absolute percecentable error.
    """
    return 100 * np.mean(np.abs((major_ts - minor_ts)/major_ts))


def smape(major_ts: np.array, minor_ts: np.array) -> float:
    """Calculate symmetrical mean absolute percecentable error between 2 metricks.

    Args:
        major_ts (np.array): original data vector.
        minor_ts (np.array): predicted data vector.

    Returns:
        float: symmetrical mean absolute percecentable error.
    """
    return 100 * np.mean(np.abs(major_ts - minor_ts)/((np.abs(minor_ts) + np.abs(major_ts))/2))


def mre(major_ts: np.array, minor_ts: np.array) -> float:
    """Calculate relative mean error between 2 metricks.

    Args:
        major_ts (np.array): original data vector.
        minor_ts (np.array): predicted data vector.

    Returns:
        float: relative mean error.
    """
    return np.mean(np.abs(major_ts - minor_ts)/np.abs(major_ts))


def rmsle(major_ts: np.array, minor_ts: np.array) -> float:
    """Calculate root mean squared logarithmic error between 2 metricks.

    Args:
        major_ts (np.array): original data vector.
        minor_ts (np.array): predicted data vector.

    Returns:
        float: root mean squared logarithmic error.
    """
    return np.sqrt(np.mean((np.log(major_ts + 1) - np.log(minor_ts + 1))**2))


def r(major_ts: np.array, minor_ts: np.array) -> float:
    """Calculate correlation coefficient.

    Args:
        major_ts (np.array): original data vector.
        minor_ts (np.array): predicted data vector.

    Returns:
        float: correlation coefficient.
    """
    return np.corrcoef(major_ts, minor_ts, rowvar=False)[0,1]


def r2(major_ts: np.array, minor_ts: np.array) -> float:
    """Calculate determination coefficient.

    Args:
        major_ts (np.array): original data vector.
        minor_ts (np.array): predicted data vector.

    Returns:
        float: determination coefficient.
    """
    return 1 - mse(major_ts = major_ts, minor_ts = minor_ts)/((major_ts - major_ts.mean())**2).mean()


def adjusted_r2(major_ts: np.array,
                              minor_ts: np.array,
                              count_features: int
                              ) -> float:
    """Calculate adjusted determination coefficient.

    Args:
        major_ts (np.array): original data vector.
        minor_ts (np.array): predicted data vector.
        count_features (int): count features in model.
        count_observe (int): count observers in ts.

    Returns:
        float: adjusted determination coefficient.
    """
    count_observe = major_ts.shape[0]
    return 1 - (mse(major_ts, minor_ts)/(count_observe - count_features))/(((major_ts - major_ts.mean())**2).mean()/(count_observe - 1))


"""File with regression model algorithm on two-stage least sqares."""
# Third Party Library
from pamm.algorithms.regression.regression_model import RegressionAlgorithm

import numpy as np

class IterationRegressionAlgorithm(RegressionAlgorithm):
        
    def _add_bias(self, x: np.array) -> np.array:
        return np.hstack([np.ones((x.shape[0], 1)),x])
    
        
    def fit(self, x: np.ndarray, y: np.ndarray) -> None:
        """Fit model.

        Args:
            x (np.ndarray): matrix of endogenous regressors (features)
            y (np.ndarray): matrix a teacher
        """
        
        remnant = y
        x = self._add_bias(x)
        rows, cols = x.shape
        tmp = np.zeros([cols,1])
        
        for col in range(1,cols):
            x_i = self._add_bias(x[:,col].reshape([rows,1]))
            beta = np.linalg.inv(x_i.T @ x_i) @ x_i.T @ remnant
            tmp[col][0] = beta[1][0]
            remnant = remnant - x_i @ beta
        
        tmp[0][0] = remnant.mean()
        self._beta = tmp[int(self._fit_bias):]
        self._bias  = tmp[0][0] * int(self._fit_bias)


# Standard Library
from abc import ABC, abstractmethod


class AbstractBaseModelAlgorithm(ABC):

    @abstractmethod
    def fit(self):
        """Get fitting of model."""

    @abstractmethod
    def predict(self):
        """Get predict of model."""
        
        
class AbstractRegressionModelAlgorithm(ABC):

    @property
    @abstractmethod
    def beta(self):
        """Get coefficients of model."""
        
    @property
    @abstractmethod
    def bias(self):
        """Get bias of model."""
        
        
class AbstractTreeModelAlgorithm(ABC):
    """Abstruct for Tree
    """
        
        
class AbstractNeyroModelAlgorithm(ABC):
    """Abstruct for Neyro
    """


"""File with regression model algorithm on two-stage least sqares."""
# Third Party Library
from pamm.algorithms.regression.regression_model import RegressionAlgorithm

import numpy as np

class StachosticGradientRegressionAlgorithm(RegressionAlgorithm):
        
    def fit(self,
        x: np.ndarray,
        y: np.ndarray,
        eta: float = .1,
        n_epochs: int = 10,
        t0: int = 1,
        t1: int = 10) -> np.ndarray:
        
        x = self._add_bias(x)
        m, n = x.shape
        theta = np.random.randn(n,1)
        for epoch in range(n_epochs):
            for i in range(m) :
                random_index = np.random.randint(m)
                xi = x[random_index:random_index + 1]
                yi = y[random_index:random_index + 1]
                gradients = 2 * xi.T @ (xi @ theta - yi)
                eta = t0/((epoch*m + i) + t1)
                theta = theta - eta * gradients
        
        self._beta = theta[int(self._fit_bias):]
        self._bias  = theta[0][0] * int(self._fit_bias)


"""File with algorithms for calculate matrix of trust data."""
# Standard Library

# Third Party Library
import numpy as np


def equal(data_table: np.ndarray) -> np.ndarray:
    return np.eye(data_table.shape[0])


def time_booster(data_table: np.ndarray) -> np.ndarray:
    return np.diag([x for x in range(0,data_table.shape[0])])


def inversely(data_table: np.ndarray) -> np.ndarray:
    return np.diag(1/data_table[:,0])


def big_target(data_table: np.ndarray) -> np.ndarray:
    return np.diag(np.abs(data_table[:,0] - np.mean(data_table)))
    

def ln_algorithm(data_table: np.ndarray) -> np.ndarray:
    return np.diag([ np.log(x) for x in range(1,data_table.shape[0]+1)])


"""File with regression model algorithm on two-stage least sqares."""
# Third Party Library
from pamm.algorithms.regression.regression_model import RegressionAlgorithm

import numpy as np

class TwoStageLeastSquaresAlgorithm(RegressionAlgorithm):
        
    def fit(self, x: np.ndarray, z:np.array, y: np.ndarray) -> None:
        """Fit model.

        Args:
            x (np.ndarray): matrix of endogenous regressors (features)
            z (np.ndarray): matrix of exogenous regressors (tools)
            y (np.ndarray): matrix a teacher
        """
        x = self._add_bias(x)
        p = z @ np.linalg.inv(z.T @ z) @ z.T
        tmp = np.linalg.inv(x.T @ p @ x) @ x.T @ p @ y
        self._beta = tmp[int(self._fit_bias):]
        self._bias  = tmp[0][0] * int(self._fit_bias)


"""File with quad prog algorithm."""
# Third Party Library
from pamm.algorithms.regression.regression_model import RegressionAlgorithm


import cvxopt as quad_prog_cvx
import numpy as np

class QPRegressionAlgorithm(RegressionAlgorithm):
    
    """Class for calculating linear model coefficients using MRLS(weighted Least Squares) + Quadprog.
    """
    def _add_bias(self, x: np.ndarray, min: np.ndarray, max: np.ndarray) -> tuple[np.ndarray]:
        if self._fit_bias:
            return (np.hstack([np.ones((x.shape[0], 1)),x]),
            np.vstack([-100000,min]),
            np.vstack([100000,max]))
        return x , min, max
    
    def fit(self,
              x: np.ndarray,
              y: np.ndarray,
              r: np.array,
              beta_min: np.ndarray,
              beta_max: np.ndarray) -> None:
        """The method of calculating coefficients is the RLS method using quadprog.

        Args:
            x_matrix (np.ndarray): matrix of regressors
            y_vector (np.ndarray): teacher vector
            r_matrix (np.ndarray): data trust matrix
            beta_min (np.atleast_2d): tuple with minimum values of regressors
            beta_max (np.atleast_2d): tuple with maximum values of regressors
        """

        x, beta_min, beta_max = self._add_bias(x, beta_min, beta_max)

        _, count_features = x.shape
        
        h = 2 * (x.T @ r @ x)
        f = (-2*(y.T @ r @ x)).T
        a = np.concatenate((np.identity(count_features), -np.identity(count_features)), axis=0)
        b = np.concatenate((beta_max, -beta_min), axis=0)

        p = quad_prog_cvx.matrix(h.astype(np.double))
        q = quad_prog_cvx.matrix(f.astype(np.double))
        g = quad_prog_cvx.matrix(a.astype(np.double))
        h = quad_prog_cvx.matrix(b.astype(np.double))

        res_quadprog = quad_prog_cvx.solvers.qp(P = p, q = q, G = g, h = h)['x'].T
        tmp = np.atleast_2d(res_quadprog)[0].reshape(count_features, 1)
        
        self._beta = tmp[int(self._fit_bias):]
        self._bias  = tmp[0][0] * int(self._fit_bias)


from  pamm.algorithms.base_model import AbstractRegressionModelAlgorithm

import numpy as np


class RegressionAlgorithm(AbstractRegressionModelAlgorithm):
    
    def __init__(self, fit_bias: bool = True) -> None:
        self._fit_bias : bool  = fit_bias
        self._beta : np.array | None = None
        self._bias : float  = 0
    
    
    def _add_bias(self, x: np.array) -> np.array:
        if self._fit_bias:
            return np.hstack([np.ones((x.shape[0], 1)),x])
        return x
    
    
    def fit(self, x: np.ndarray, y: np.ndarray) -> None:
        """Fit model.
        Args:
            x_matrix (np.ndarray): matrix of regression
            y_vector (np.ndarray): matrix a teacher data
        """
        x = self._add_bias(x)
        tmp = np.linalg.inv(x.T @ x) @ x.T @ y
        self._beta = tmp[int(self._fit_bias):]
        self._bias  = tmp[0][0] * int(self._fit_bias)


    def predict(self, x: np.ndarray) -> np.ndarray:
        """Get predict values.

        Args:
            x_matrix (np.ndarray): matrix of regression

        Returns:
            np.ndarray: Predicted values
        """
        return x @ self._beta + self._bias

    @property
    def beta(self) -> np.ndarray:
        """Get beta coefficents.

        Returns:
            np.ndarray: vector of beta coefficients.
        """
        return self._beta

    @beta.setter
    def beta(self, new_beta: tuple) -> None:
        """Set beta coefficients in manual mode.

        Args:
            new_beta (tuple): new beta coefficients
        """
        self._beta = np.atleast_2d(new_beta).T

    @property
    def bias(self) -> float:
        """Get intercept.

        Returns:
            float: intercept
        """
        return self._bias

    @bias.setter
    def bias(self, num: float) -> None:
        """Set intercept a manual mode.

        Args:
            intercept (float): new intercept
        """
        self._bias = num


import numpy as np

def bgd(x: np.ndarray,
        y: np.ndarray,
        eta: float = .1,
        max_iter: int = 1000) -> np.ndarray:
    m, n = x.shape
    theta = np.random.randn(n,1)
    for iteration in range(max_iter): 
        gradients = 2/m * x.T @ (x @ theta - y) 
        theta = theta - eta * gradients
    return theta



def sgd(x: np.ndarray,
        y: np.ndarray,
        eta: float = .1,
        n_epochs: int = 10,
        t0: int = 1,
        t1: int = 10) -> np.ndarray:
    m, n = x.shape
    theta = np.random.randn(n,1)
    for epoch in range(n_epochs):
        for i in range(m) :
            random_index = np.random.randint (m)
            xi = x[random_index:random_index + 1]
            yi = y[random_index:random_index + 1]
            gradients = 2 * xi.T @ (xi @ theta - yi)
            eta = t0/((epoch*m + i) + t1)
            theta = theta - eta * gradients
    return theta


"""File with ridge regression model algorithm."""
# Third Party Library
from pamm.algorithms.regression.regression_model import RegressionAlgorithm

import numpy as np

class LassoRegressionAlgorithm(RegressionAlgorithm):
        
        
    def _sign(self, num: float) -> float:
        if num == 0: return num
        if num > 0: return 1
        if num < 0: return -1
    
    def _sign_vectorize(self, b: np.ndarray) -> np.ndarray:
        vector = np.zeros(b.shape)
        for x in range(b.shape[0]):
            vector[x,0] = self._sign(b[x,0])
        return vector
    
    def fit(self,
        x: np.ndarray,
        y: np.ndarray,
        alpha: float = .01,
        eta: float = .1,
        n_epochs: int = 10,
        t0: int = 1,
        t1: int = 10) -> np.ndarray:
        
        x = self._add_bias(x)
        m, n = x.shape
        b = np.random.randn(n,1)
        for epoch in range(n_epochs):
            for i in range(m) :
                random_index = np.random.randint (m)
                xi = x[random_index:random_index + 1]
                yi = y[random_index:random_index + 1]
                gradients = 2 * xi.T @ (xi @ b - yi) + alpha * self._sign_vectorize(b)
                eta = t0/((epoch*m + i) + t1)
                b = b - eta * gradients
                
        self._beta = b[int(self._fit_bias):]
        self._bias  = b[0][0] * int(self._fit_bias)


"""File with functions for read ts from files"""

# Standart library
import os
import asyncio
import datetime
import numpy as np
import pandas as pd
import openpyxl as op
from io import StringIO
from xlsx2csv import Xlsx2csv

# delete void columns from data
def del_void_columns(df: pd.DataFrame) -> pd.DataFrame:
    void_columns = [col for col in df.columns if pd.isnull(pd.to_numeric(df[col], errors='coerce').mean())]
    return df.drop(void_columns ,axis=1)

# function for take the name of files in directory
def generate_files_list(folder_path :str) -> list[str]: 
    return [os.path.join(folder_path, f) for f in os.listdir(folder_path) if os.path.isfile(os.path.join(folder_path, f))]

# function for search index with datetime
def search_date_index(df: pd.DataFrame) -> list[int]:
    date_index = [num for num, col in enumerate(df) if isinstance(np.datetime64('2005-02', 'D'), df[col].dtype.type)]
    date_index.append(len(df.columns)+1)
    return date_index

# function for generate data frames with ones dates
def generate_df(data_frame: pd.DataFrame) -> list[pd.DataFrame]:
    dates = search_date_index(data_frame)

    list_df = []

    for num, x in enumerate(dates):
        if len(dates) == num + 1:
            break
        list_df.append(data_frame.iloc[:, dates[num]:dates[num+1]])
    return list_df

# function for generate timeseries
def generate_ts(data_frame: pd.DataFrame) -> dict:
    data_frames = generate_df(data_frame)
    dict_ts = {}
    for df in data_frames:
        for num,col in enumerate(df.columns):
            if num != 0:
                temp_df = df.iloc[:,[0,num]]
                temp_df[col] = pd.to_numeric(data_frame[col], errors='coerce')
                dict_ts[col] = temp_df.dropna().to_numpy()
                dict_ts[col][:,0] = dict_ts[col][:,0].astype(np.datetime64)
    return dict_ts

# function for read excel files
def read_excel(path: str) -> dict[str : np.array]:
    data = {}
    file = pd.ExcelFile(path)
    for sheet in file.sheet_names:
        tmp = del_void_columns(pd.read_excel(file, sheet_name=sheet))
        data.update(generate_ts(tmp))
    return data


# function for read csv files
def read_csv(path: str) -> None:
    tmp = del_void_columns(pd.read_csv(path))
    data = generate_ts(tmp)

# dictionary with file-extentions : function for read
EXTENTION_DICT = {
    '.xlsx':read_excel,
    '.xlsm':read_excel,
    '.xlsb':read_excel,
    '.xlam':read_excel,
    '.xltx':read_excel,
    '.xltm':read_excel,
    '.xls':read_excel,
    '.xla':read_excel,
    '.xlt':read_excel,
    '.csv':read_csv,
    '.odf':read_excel
}

# function for read all tables-like files in directory
def read_directory(path: str) -> None:
    data = {}
    file_names = generate_files_list(path)
    for filename in file_names:
        _, fileExtension = os.path.splitext(filename)
        if fileExtension in EXTENTION_DICT:
            tempFrame = EXTENTION_DICT[fileExtension](filename)
            data.update(tempFrame)
    return data


from pamm.algorithms.base_model import AbstractBaseModelAlgorithm

import numpy as np

def cv(model: AbstractBaseModelAlgorithm, x: np.array, y: np.array, cv: int = 2) -> None:
    len_ds = int(np.ceil(x.shape[0]/cv))
    models = [model for _ in range(cv)]
    
    for num, model in enumerate(models):
        x_reduce = x[num*len_ds:(num+1)*len_ds]
        y_reduce = y[num*len_ds:(num+1)*len_ds]
        model.fit(x = x_reduce, y = y_reduce)


"""File with ridge regression model algorithm."""
# Third Party Library
from pamm.algorithms.regression.regression_model import RegressionAlgorithm

import numpy as np

class RidgeRegressionAlgorithm(RegressionAlgorithm):
        
    def fit(self, x: np.ndarray, y: np.ndarray, alpha: float = .1) -> None:
        """Fit model.

        Args:
            x (np.ndarray): matrix of regressors
            y (np.ndarray): matrix a teacher
            alpha (float): regularization parameter
        """
        x = self._add_bias(x)
        I = np.eye(x.shape[1])
        tmp = np.linalg.inv(x.T @ x + alpha * I) @ x.T @ y
        self._beta = tmp[int(self._fit_bias):]
        self._bias  = tmp[0][0] * int(self._fit_bias)



from pamm.algorithms.regression.regression_model import RegressionAlgorithm
from pamm.algorithms.regression.tsls_model import TwoStageLeastSquaresAlgorithm
from pamm.algorithms.regression.jackknife_model import JackKnifeRegressionAlgorithm
from pamm.algorithms.regression.iterations_regression import IterationRegressionAlgorithm
from pamm.algorithms.regression.regularization.ridge_model import RidgeRegressionAlgorithm
from pamm.algorithms.regression.regularization.lasso_model import LassoRegressionAlgorithm
from pamm.algorithms.regression.regularization.elastic_net import ElasticNetRegressionAlgorithm
from pamm.algorithms.regression.wls.wls_model import WeigthedRegressionAlgorithm
from pamm.algorithms.regression.wls.wtsls_model import WeightedTwoStageLeastSquaresAlgorithm
from pamm.algorithms.regression.wls.mls_model import ModifyRegressionAlgorithm
from pamm.algorithms.regression.wls.qp_model import QPRegressionAlgorithm
from pamm.algorithms.regression.gradient_boost.pgd_regression import BatchGradientRegressionAlgorithm
from pamm.algorithms.regression.gradient_boost.sgd_regression import StachosticGradientRegressionAlgorithm


"""File with regression model algorithm on two-stage least sqares."""
# Third Party Library
from  pamm.algorithms.regression.regression_model import RegressionAlgorithm

import numpy as np

class JackKnifeRegressionAlgorithm(RegressionAlgorithm):
        
    def fit(self, x: np.ndarray, y: np.ndarray) -> None:
        """Fit model.

        Args:
            x (np.ndarray): matrix of endogenous regressors (features)
            y (np.ndarray): matrix a teacher
        """
        x = self._add_bias(x)
        
        betas = np.zeros(x.T.shape)
        
        for i in range(x.shape[0]):
            x_reduce = np.delete(x, i, axis=0)
            y_reduce = np.delete(y, i, axis=0)
            betas[:,0] = (np.linalg.inv(x_reduce.T @ x_reduce) @ x_reduce.T @ y_reduce).T
        
        tmp = np.atleast_2d(np.mean(betas, axis=1)).T
        
        self._beta = tmp[int(self._fit_bias):]
        self._bias  = tmp[0][0] * int(self._fit_bias)


"""File with algorithms for calculate matrix of trust beta coefficients."""
# Standard Library
from abc import ABC, abstractmethod

# Third Party Library
import numpy as np


def max_truth(data_table: np.ndarray) -> np.ndarray:
    size_to_append = 1000000
    beta_vector = np.where(np.abs(data_table.T) > 0, size_to_append, 0)[0]
    return np.diag(beta_vector)


"""File with methods for generate list algorithms."""

# Third Party Library
import numpy as np
from pamm.algorithms.regression.wls.weights import weights_data_algorithms
from pamm.algorithms.regression.wls.weights import weights_beta_algorithms 
from inspect import getmembers, isfunction
from collections.abc import Sequence


all_weights_data_algorithm = {func[0]:func[1] for func in getmembers(weights_data_algorithms, isfunction)}
all_weights_beta_algorithm = {func[0]:func[1] for func in getmembers(weights_beta_algorithms, isfunction)}

def apply_weights_data_algorithm(
    data: np.ndarray,
    algorithm: Sequence[list(all_weights_data_algorithm)]  = 'usual_algorithm'
    ) -> np.ndarray:

    return all_weights_data_algorithm[algorithm](data)


def apply_weights_beta_algorithm(
    data: np.ndarray,
    algorithm: Sequence[list(all_weights_beta_algorithm)]  = 'usual_algorithm'
    ) -> np.ndarray:

    return all_weights_beta_algorithm[algorithm](data)



"""File with cross-validation regression model algorithm."""
# Third Party Library
import numpy as np
from pamm.algorithms.base_model import AbstractRegressionAlgorithm


class CVRegressionAlgorithm(AbstractRegressionAlgorithm):
    
    def __init__(self, fit_intercept: bool = True) -> None:
        self._fit_intercept = fit_intercept

    def fit(self, x_matrix: np.ndarray, y_vector: np.ndarray, cv: int = 5) -> None:
        """Fit model.

        Args:
            x_matrix (np.ndarray): matrix of regression
            y_vector (np.ndarray): matrix a teacher data
            cv (int): the number of divisions of the dataset
        """

        if self._fit_intercept:
            x_matrix = np.hstack([np.ones((x_matrix.shape[0], 1)),x_matrix])

        len_ds = int(np.ceil(x_matrix.shape[0]/cv))

        b_matrix = np.zeros((len_ds-1,x_matrix.shape[1]))
        
        for ds in range(cv):
            x_reduce = x_matrix[ds*len_ds:(ds+1)*len_ds]
            y_reduce = y_vector[ds*len_ds:(ds+1)*len_ds]
            beta = np.linalg.inv(x_reduce.T @ x_reduce) @ x_reduce.T @ y_reduce
            b_matrix[ds] = beta.T

        tmp = np.atleast_2d(np.mean(b_matrix, axis=0)).T
        
        self._beta_coefs = tmp[int(self._fit_intercept):]
        self._intercept  = tmp[0][0] * int(self._fit_intercept)

    def predict(self, x_matrix: np.ndarray) -> np.ndarray:
        """Get predict values.

        Args:
            x_matrix (np.ndarray): matrix of regression

        Returns:
            np.ndarray: Predicted values
        """
        return x_matrix @ self._beta_coefs + self._intercept

    @property
    def beta_coefs(self) -> np.ndarray:
        """Get beta coefficents.

        Returns:
            np.ndarray: vector of beta coefficients.
        """
        return self._beta_coefs

    @beta_coefs.setter
    def beta_coefs(self, new_beta: tuple) -> None:
        """Set beta coefficients in manual mode.

        Args:
            new_beta (tuple): new beta coefficients
        """
        self._beta_coefs = np.atleast_2d(new_beta).T

    @property
    def intercept(self) -> float:
        """Get intercept.

        Returns:
            float: intercept
        """
        return self._intercept

    @intercept.setter
    def intercept(self, intercept: float) -> None:
        """Set intercept a manual mode.

        Args:
            intercept (float): new intercept
        """
        self._intercept = intercept


import datetime
import numpy as np
import datetime

def side2_percentile_filter(ts: np.ndarray, percentile: float) -> np.ndarray:
    values = ts[:,1]
    up_percentile = np.percentile(values,percentile)
    low_percentile = np.percentile(values,100 - percentile)

    mask = (values < up_percentile) ,(values > low_percentile)
    return ts[np.logical_and(*mask)]

def set_delay(ts: np.ndarray, delay:datetime.timedelta) -> np.ndarray:
    data = ts.copy()
    data[:,0] = data[:,0] + delay
    return data



from abc import ABC
import pandas as pd
import numpy as np
from typing import Union, Optional
from collections.abc import Iterable

class DataSample(ABC):
    
    def __init__(self, data: Union[dict, pd.DataFrame], y_column: str, edge: Union[int, float] = .3, time_name: str = 'Time'):

        self._data = data
        self._y = self._validate_y(y_column)
        self._edge = self._validate_edge(edge)
        self._time = time_name


    def __len__(self) -> int:
        return len(self._data[self._y])


    def __getitem__(self, col:str) -> np.ndarray:
        return self._data[col]


    def _percent2index(self):
        delimeter = int(np.ceil(len(self.data) * (100 - self._precent)/100))


    def _validate_edge(self,num) -> int:
        if num < 0:
            raise ValueError("Count values must be positve only.")
        if isinstance(num, int):
            if num > len(self._data[self._y]):
                raise ValueError(f"The value must be less than count of experiment: ({len(self._data[self._y])}).")
            return num
        if num > 1:
            raise ValueError("The value defined as a percentage must be less than 1.")
        return int(np.ceil(len(self._data[self._y]) * (1-num)))
    

    def _validate_y(self,y_column) -> int:
        if y_column in list(self._data.keys()):
            return y_column
        else:
            raise ValueError(f"The y_column name must be contained in the data table: {list(self._data.keys())}.")


    def _matrix(self, names: Union[None ,str, Iterable[str]], up: int, down: int) -> tuple[np.ndarray]:
        _res = None
        if isinstance(names,str):
            names = [names]
        elif names is None:
            names = [name for name in self._data.keys() if name not in [self._y, self._time]]
        for num, name in enumerate(names):
            if num == 0:
                _res = self._data.get(name)
            else:
                _res = np.hstack((_res, self._data.get(name)))
        return (_res[down:up],
        self._data[self._y][down:up],
        self._data[self._time][down:up])

    
    def train(self, names: Union[None, str, Iterable] = None) -> tuple[np.ndarray]:
            return self._matrix(names = names, up = self._edge, down = 0)


    def test(self, names: Union[None, str, Iterable] = None) -> tuple[np.ndarray]:
            return self._matrix(names = names, up = len(self), down = self._edge)


    @property
    def edge(self) -> int:
        return self._edge


    @property
    def data(self) -> dict:
        return self._data


    @edge.setter
    def edge(self, edge:Union[int, float]) -> None:
        self._edge = self._validate_edge(edge)


    @property
    def y(self) -> str:
        return self._y
    
    @property
    def to_DataFrame(self):
        all_names = [name for name in self._data.keys()]
        tmp = self._matrix(names = all_names, up = len(self), down = 0)[0]
        return pd.DataFrame(data = tmp, columns = all_names)


"""File with merge algorithms for creation dataset"""

# Standart library
import datetime
import numpy as np
import pandas as pd


# Function for converte ts to pandas.DataFreme
def ts2frame(ts :np.ndarray, ts_name: str) -> pd.DataFrame:
    df = pd.DataFrame(data= ts, columns= ['Time', ts_name])
    return df
    

# Function for merge two datafeme
def merge(major_df: pd.DataFrame, minor_df:pd.DataFrame, method: str = 'nearest', tol: str = '1000000h'):
    return pd.merge_asof(major_df, minor_df, on=0, tolerance= pd.Timedelta(tol), direction = method)


# Function for merge all ts
def merge_all(data_dict: dict, sync_name: str, config: dict, max_tol: str = '1000000h', default_method: str = 'nearest'):
    data_frame = pd.DataFrame(data_dict[sync_name])
    dict_temp = data_dict.copy()
    del dict_temp[sync_name]
    for key, value in dict_temp.items():
        temp = pd.DataFrame(value)
        try:
            data_frame = merge(data_frame, temp, config[key]["method"], config[key]["tol"])
        except:
            data_frame = merge(data_frame, temp, default_method, max_tol)
    
    data_frame = data_frame.dropna()
    res = {} 
    for num, name in enumerate(['Time', sync_name] + [x for x,y in dict_temp.items() ]):
        res[name] = np.atleast_2d(data_frame.iloc[:,num]).T
    
    res['Time'] = res['Time'].astype(datetime.datetime)

    return res


"""File with weigthed regression model algorithm."""
# Third Party Library
from pamm.algorithms.regression.regression_model import RegressionAlgorithm

import numpy as np

class WeigthedRegressionAlgorithm(RegressionAlgorithm):
        
    def fit(self, x: np.ndarray, y: np.ndarray, w: np.ndarray) -> None:
        """Fit model.

        Args:
            x (np.ndarray): matrix of regression
            y (np.ndarray): matrix a teacher data
            w (np.ndarray): matrix of weights
        """
        x = self._add_bias(x)
        tmp = np.linalg.inv(x.T @ w @ x) @ x.T @ w @ y
        self._beta = tmp[int(self._fit_bias):]
        self._bias  = tmp[0][0] * int(self._fit_bias)


"""File with regression model algorithm on two-stage least sqares."""
# Third Party Library
from pamm.algorithms.regression.regression_model import RegressionAlgorithm

import numpy as np

class BatchGradientRegressionAlgorithm(RegressionAlgorithm):
        
    def fit(self,
        x: np.ndarray,
        y: np.ndarray,
        eta: float = .1,
        max_iter: int = 1000) -> np.ndarray:
        
        x = self._add_bias(x)
        m, n = x.shape
        theta = np.random.randn(n,1)
        for iteration in range(max_iter): 
            gradients = 2/m * x.T @ (x @ theta - y) 
            theta = theta - eta * gradients
        
        self._beta = theta[int(self._fit_bias):]
        self._bias  = theta[0][0] * int(self._fit_bias)